"""Executor test

This file provides tests for the Executor of uppaal2jetracer.
"""

from uppaal2jetracer.controller.executor import Executor, ExecutionError
from uppaal2jetracer.uppaalmodel.frame import Frame
from uppaal2jetracer.uppaalmodel.variable import Array, IntVariable
from uppaal2jetracer.declarations.declarations_ast import ID, Return, Constant, IdentifierType, \
    BinaryOp, UnaryOp, Decl, TypeDecl, Assignment, Compound, FuncDecl, ParamList, FuncDef, \
    FuncCall, ExprList, ArrayDecl, InitList, ArrayRef, Node

frame: Frame = Frame()

executor: Executor = Executor(frame)


def test_id_type():
    """
    Test that the executor visits an IdentifierType node correctly.
    Example: int
    """

    type_node: IdentifierType = IdentifierType("int")
    assert type_node.accept(executor) == "int"

    _assert_execution_error(IdentifierType(None), "Identifier type has no type.")


def test_constant():
    """
    Test that the executor visits a Constant node correctly.
    Example: 9
    """

    constant_node: Constant = Constant("double", 9.12)
    assert constant_node.accept(executor) == 9.12

    _assert_execution_error(Constant(None, 7), "Constant has no type.")
    _assert_execution_error(Constant("double", None), "Constant has no value.")
    _assert_execution_error(Constant("type", 7), "Constant type 'type' is not supported.")


def test_variable_decl():
    """
    Test that the executor visits a Decl node correctly.
    Example: int x = 20; int y = 10;
    """

    x_type_node: IdentifierType = IdentifierType("int")
    y_type_node: IdentifierType = IdentifierType("int")
    x_type_decl_node: TypeDecl = TypeDecl("x", [], x_type_node)
    y_type_decl_node: TypeDecl = TypeDecl("y", [], y_type_node)
    x_value: Constant = Constant("int", 20)
    y_value: Constant = Constant("int", 10)
    x_decl: Decl = Decl("x", [], None, None, x_type_decl_node, x_value, None)
    y_decl: Decl = Decl("y", [], None, None, y_type_decl_node, y_value, None)
    x_decl.accept(executor)
    y_decl.accept(executor)
    assert isinstance(frame.get_variable("x"), IntVariable)
    assert isinstance(frame.get_variable("y"), IntVariable)
    assert frame.get_variable("x").value == 20
    assert frame.get_variable("y").value == 10

    _assert_execution_error(TypeDecl("x", [], None), "Type declaration has no type.")
    _assert_execution_error(TypeDecl("x", None, x_type_node),
                            "Type declaration has no qualifier list.")
    _assert_execution_error(Decl("x", [], None, None, None, None, None),
                            "Declaration has no type.")
    _assert_execution_error(Decl(None, None, None, None, x_type_decl_node, x_value, None),
                            "Declaration has no name.")


def test_id():
    """
    Test that the executor visits an ID node correctly.
    Example: x
    """

    id_node: ID = ID("x")
    assert id_node.accept(executor) == 20

    _assert_execution_error(ID(None), "Identifier has no name.")


def test_array_decl():
    """
    Test that the executor visits a ArrayDecl node correctly.
    Example: int a[2] = {x, y};
    """

    x_id_node: ID = ID("x")
    y_id_node: ID = ID("y")
    init_list_node: InitList = InitList([x_id_node, y_id_node])

    ar_type_node: IdentifierType = IdentifierType("int")
    constant_node: Constant = Constant("int", 2)
    ar_array_decl_node: ArrayDecl = ArrayDecl(ar_type_node, constant_node, ["const"])
    ar_decl: Decl = Decl("ar", ["const"], None, None, ar_array_decl_node,
                         init_list_node, None)

    ar_decl.accept(executor)
    assert isinstance(frame.get_variable("ar"), Array)
    assert frame.get_variable("ar").is_constant

    _assert_execution_error(ArrayDecl(None, None, None),
                            "Array declaration has no dimensional qualifier.")


def test_array_ref():
    """
    Test that the executor visits a ArrayRef node correctly.
    Example: a[0]
    """

    constant_node: Constant = Constant("int", 0)
    array_ref_node: ArrayRef = ArrayRef("ar", constant_node)

    ar = frame.get_variable("ar")
    assert isinstance(ar, Array)
    assert array_ref_node.accept(executor) == ar.get_index(0)

    _assert_execution_error(ArrayRef("a", constant_node), "Array 'a' is not declared.")
    _assert_execution_error(ArrayRef("x", constant_node), "Variable 'x' is not an array.")
    _assert_execution_error(ArrayRef("a", None), "Array reference to 'a' has no index.")


def test_unary_op():
    """
    Test that the executor visits an UnaryOP node correctly.
    Example: x--
    """

    id_node: ID = ID("x")
    unary_op_node: UnaryOp = UnaryOp("p--", id_node)
    assert unary_op_node.accept(executor) == frame.get_variable("x").value - 1

    _assert_execution_error(UnaryOp("p--", None), "Unary operation has no expression.")
    _assert_execution_error(UnaryOp("+10", None), "Unary operator '+10' is not supported.")


def test_binary_op():
    """
    Test that the executor visits a BinaryOp node correctly.
    Example: x + y
    """

    x_id_node: ID = ID("x")
    y_id_node: ID = ID("y")
    binary_op_node: BinaryOp = BinaryOp("+", x_id_node, y_id_node)
    assert (binary_op_node.accept(executor) == frame.get_variable("x").value
            + frame.get_variable("y").value)

    binary_op_node = BinaryOp("/", x_id_node, y_id_node)
    assert (binary_op_node.accept(executor) == frame.get_variable("x").value
            / frame.get_variable("y").value)

    _assert_execution_error(BinaryOp("/", x_id_node, None),
                            "Binary operation has no right expression.")
    _assert_execution_error(BinaryOp("/", None, None), "Binary operation has no left expression.")
    _assert_execution_error(BinaryOp("=", x_id_node, y_id_node), "Binary operator '=' is not supported.")


def test_assignment():
    """
    Test that the executor visits an Assignment node correctly.
    Example: x = y + 3;
    """

    x_id_node: ID = ID("x")
    y_id_node: ID = ID("y")
    constant_node: Constant = Constant("int", 3)
    binary_op_node: BinaryOp = BinaryOp("+", y_id_node, constant_node)
    assignment_node: Assignment = Assignment("=", x_id_node, binary_op_node)
    assignment_node.accept(executor)
    assert isinstance(frame.get_variable("x"), IntVariable)
    assert isinstance(frame.get_variable("y"), IntVariable)
    assert frame.get_variable("x").value == frame.get_variable("y").value + 3
    assert frame.get_variable("y").value == 10

    _assert_execution_error(Assignment("=", x_id_node, None), "Assignment has no right expression.")
    _assert_execution_error(Assignment("=", None, None), "Assignment has no left expression.")
    _assert_execution_error(Assignment("?", x_id_node, y_id_node),
                            "Assignment operator '?' is not supported.")


def test_return():
    """
    Test that the executor visits a Return node correctly.
    Example: return x + y;
    """

    x_id_node: ID = ID("x")
    y_id_node: ID = ID("y")
    binary_op_node: BinaryOp = BinaryOp("+", x_id_node, y_id_node)
    return_node: Return = Return(binary_op_node)
    assert (return_node.accept(executor) == frame.get_variable("x").value
            + frame.get_variable("y").value)

    _assert_execution_error(Return(None, None), "Return has no expression.")


def test_func_def():
    """
    Test that the executor visits a FuncDef node correctly.
    Example: int add(int a, int b) {return a + b;}
    """

    a_type_node: IdentifierType = IdentifierType("int")
    b_type_node: IdentifierType = IdentifierType("int")
    a_type_decl_node: TypeDecl = TypeDecl("a", [], a_type_node)
    b_type_decl_node: TypeDecl = TypeDecl("b", [], b_type_node)
    a_decl: Decl = Decl("a", [], None, None, a_type_decl_node, None, None)
    b_decl: Decl = Decl("b", [], None, None, b_type_decl_node, None, None)
    add_params_node: ParamList = ParamList([a_decl, b_decl])
    add_type_node: IdentifierType = IdentifierType("int")
    add_type_decl_node: TypeDecl = TypeDecl("add", [], add_type_node)
    add_func_decl_node: FuncDecl = FuncDecl(add_params_node, add_type_decl_node)
    add_decl_node: Decl = Decl("add", [], None, None, add_func_decl_node, None, None)

    a_id_node: ID = ID("a")
    b_id_node: ID = ID("b")
    binary_op_node: BinaryOp = BinaryOp("+", a_id_node, b_id_node)
    return_node: Return = Return(binary_op_node)
    compound_node: Compound = Compound([return_node])

    add_func_def_node: FuncDef = FuncDef(add_decl_node, None, compound_node)
    add_func_def_node.accept(executor)
    assert isinstance(frame.get_function("add"), FuncDef)
    assert frame.get_variable("a") is None
    assert frame.get_variable("b") is None

    _assert_execution_error(FuncDecl(None, None),
                            "Function declarations are supported but should never be visited.")
    _assert_execution_error(FuncDef(None, None, None), "Function definition has no declaration.")

def test_func_call():
    """
    Test that the executor visits a FuncCall node correctly.
    Example: add(x, y)
    """

    add_id_node: ID = ID("add")
    x_id_node: ID = ID("x")
    y_id_node: ID = ID("y")
    add_args_node: ExprList = ExprList([x_id_node, y_id_node])
    add_func_call: FuncCall = FuncCall(add_id_node, add_args_node)

    assert (add_func_call.accept(executor) == frame.get_variable("x").value
            + frame.get_variable("y").value)

    _assert_execution_error(FuncCall(add_id_node, None),
                            "Function call 'add' provides an invalid amount of arguments.")
    _assert_execution_error(FuncCall(ID("some_func"), None), "Function 'some_func' is not declared.")


def _assert_execution_error(node: Node, message: str):
    try:
        node.accept(executor)
        assert False
    except ExecutionError as e:
        assert e.args[0] == message
