"""
Test cases for uppaal controller.
"""
from uppaal2jetracer.declarations.declarations_ast import ID, Constant, BinaryOp, \
    Assignment, ArrayRef

from uppaal2jetracer.controller.uppaal_controller import AutomatonController, SystemController
from uppaal2jetracer.uppaalmodel.system import Automaton, System
from uppaal2jetracer.uppaalmodel.frame import Frame
from uppaal2jetracer.uppaalmodel.elements import Location, Transition, Guard, Invariant, Update, \
    Synchronization
from uppaal2jetracer.uppaalmodel.variable import IntVariable, Channel, Array


def test_can_step_guard():
    """
    Test the can step functionality of the automaton_controller.
    The test transition has a guard.
    """
    val1 = ID("x")
    val2 = Constant("int", 4)
    bin_op = BinaryOp(">", val1, val2)
    guard = Guard(bin_op)

    frame = Frame()
    value = IntVariable(False, 4)
    frame.add_binding("x", value)

    location1 = Location("id0", "location1")
    location2 = Location("id1", "location2")
    transition = Transition(guard = guard)
    transition.source = location1
    transition.target = location2
    location1.add_transition(transition)

    automaton = Automaton("automaton", frame, [location1, location2], location1)
    automaton_controller = AutomatonController(automaton)

    assert not automaton_controller.can_step(transition)
    value.value = 5
    assert automaton_controller.can_step(transition)


def test_can_step_invariant():
    """
    Test the can step functionality of the automaton_controller.
    The target location of the test transition has an invariant.
    """
    val1 = ID("x")
    val2 = Constant("int", 4)
    bin_op = BinaryOp(">", val1, val2)
    invariant = Invariant(bin_op)

    frame = Frame()
    value = IntVariable(False, 4)
    frame.add_binding("x", value)

    location1 = Location("id0", "location1")
    location2 = Location("id1", "location2", invariant)
    transition = Transition()
    transition.source = location1
    transition.target = location2
    location1.add_transition(transition)

    automaton = Automaton("automaton", frame, [location1, location2], location1)
    automaton_controller = AutomatonController(automaton)

    assert not automaton_controller.can_step(transition)
    value.value = 5
    assert automaton_controller.can_step(transition)


def test_can_step_update_influences_invariant():
    """
    Test the can step functionality of the automaton_controller.
    The test transition has an update that influences the result of the target invariant.
    """
    val1 = ID("x")
    val2 = Constant("int", 4)
    val3 = Constant("int", 5)
    bin_op = BinaryOp(">", val1, val2)
    assignment = Assignment("=", val1, val3)
    invariant = Invariant(bin_op)
    update = Update([assignment])

    frame = Frame()
    value = IntVariable(False, 4)
    frame.add_binding("x", value)

    location1 = Location("id0", "location1")
    location2 = Location("id1", "location2", invariant)
    transition = Transition(update = update)
    transition.source = location1
    transition.target = location2
    location1.add_transition(transition)

    automaton = Automaton("automaton", frame, [location1, location2], location1)
    automaton_controller = AutomatonController(automaton)

    assert automaton_controller.can_step(transition)
    assert frame.get_variable("x").value == 4


def test_execute_sync_value():
    """
    Test execution of synchronization nodes where the synchronization is a value.
    """
    val1 = ID("x")
    frame = Frame()
    chan = Channel(False)
    frame.add_binding("x", chan)

    location1 = Location("id0", "location1")
    synchronization = Synchronization(True, val1)
    automaton = Automaton("automaton", frame, [location1], location1)
    automaton_controller = AutomatonController(automaton)

    resulting_channel = automaton_controller.execute_sync(synchronization)
    assert resulting_channel == chan


def test_execute_sync_array():
    """
    Test execution of synchronization nodes where the synchronization is an array.
    """
    val1 = Constant("int", 0)
    val2 = ArrayRef("x", val1)
    frame = Frame()
    chan = Channel(False)
    array = Array(False, [chan])
    frame.add_binding("x", array)

    location1 = Location("id0", "location1")
    synchronization = Synchronization(True, val2)
    automaton = Automaton("automaton", frame, [location1], location1)
    automaton_controller = AutomatonController(automaton)

    resulting_channel = automaton_controller.execute_sync(synchronization)
    assert resulting_channel == chan


def test_execute_step_location_change():
    """
    Test execute step method if the current location changes.
    """
    frame = Frame()

    location1 = Location("id0", "location1")
    location2 = Location("id1", "location2")
    transition = Transition()
    transition.source = location1
    transition.target = location2
    location1.add_transition(transition)

    automaton = Automaton("automaton", frame, [location1, location2], location1)
    automaton_controller = AutomatonController(automaton)

    assert automaton_controller.current_location == location1

    automaton_controller.execute_step(transition)

    assert automaton_controller.current_location == location2


def test_execute_step_update():
    """
    Test execute step if the update is executed.
    """
    val1 = ID("x")
    val2 = Constant("int", 100)
    assignment = Assignment("=", val1, val2)
    update = Update([assignment])
    frame = Frame()
    val = IntVariable(False, 5)
    frame.add_binding("x", val)

    location1 = Location("id0", "location1")
    location2 = Location("id1", "location2")
    transition = Transition(update = update)
    transition.source = location1
    transition.target = location2
    location1.add_transition(transition)

    automaton = Automaton("automaton", frame, [location1, location2], location1)
    automaton_controller = AutomatonController(automaton)

    assert frame.get_variable("x").value == 5

    automaton_controller.execute_step(transition)

    assert frame.get_variable("x").value == 100


def test_step_no_sync():
    """
    Test step method with transition with no synchronization.
    """
    frame = Frame()

    location1 = Location("id0", "location1")
    location2 = Location("id1", "location2")
    transition = Transition()
    transition.source = location1
    transition.target = location2
    location1.add_transition(transition)

    automaton = Automaton("automaton", frame, [location1, location2], location1)
    automaton_controller = AutomatonController(automaton)

    assert automaton_controller.step() == (transition, None)


def test_step_no_transition():
    """
    Test step method with no transition.
    """
    frame = Frame()
    location = Location("id0", "location")

    automaton = Automaton("automaton", frame, [location], location)
    automaton_controller = AutomatonController(automaton)

    assert automaton_controller.step() == (None, None)


def test_step_sending_broadcast_sync():
    """
    Test step method with transition with sending broadcast transition.
    """
    frame = Frame()
    chan = Channel(True)
    frame.add_binding("x", chan)
    val = ID("x")
    synchronization = Synchronization(True, val)

    location1 = Location("id0", "location1")
    location2 = Location("id1", "location2")
    transition = Transition(synchronization = synchronization)
    transition.source = location1
    transition.target = location2
    location1.add_transition(transition)

    automaton = Automaton("automaton", frame, [location1, location2], location1)
    automaton_controller = AutomatonController(automaton)

    assert automaton_controller.step() == (transition, chan)


def test_step_receiving_broadcast_sync():
    """
    Test step method with transition with receiving broadcast transition.
    """
    frame = Frame()
    chan = Channel(True)
    frame.add_binding("x", chan)
    val = ID("x")
    synchronization = Synchronization(False, val)

    location1 = Location("id0", "location1")
    location2 = Location("id1", "location2")
    transition = Transition(synchronization = synchronization)
    transition.source = location1
    transition.target = location2
    location1.add_transition(transition)

    automaton = Automaton("automaton", frame, [location1, location2], location1)
    automaton_controller = AutomatonController(automaton)

    assert automaton_controller.step() == (None, None)


def test_step_sending_binary_sync_has_no_pair():
    """
    Test step method with transition with sending binary transition without a pair.
    """
    frame = Frame()
    chan = Channel(False)
    frame.add_binding("x", chan)
    val = ID("x")
    synchronization = Synchronization(True, val)

    location1 = Location("id0", "location1")
    location2 = Location("id1", "location2")
    transition = Transition(synchronization = synchronization)
    transition.source = location1
    transition.target = location2
    location1.add_transition(transition)

    automaton = Automaton("automaton", frame, [location1, location2], location1)
    automaton_controller = AutomatonController(automaton)

    assert automaton_controller.step() == (None, None)
    assert chan.sending_transitions[0] == transition
    assert len(automaton_controller._clean_up_list) == 1


def test_step_receiving_binary_sync_has_no_pair():
    """
    Test step method with transition with receiving binary transition without a pair.
    """
    frame = Frame()
    chan = Channel(False)
    frame.add_binding("x", chan)
    val = ID("x")
    synchronization = Synchronization(False, val)

    location1 = Location("id0", "location1")
    location2 = Location("id1", "location2")
    transition = Transition(synchronization = synchronization)
    transition.source = location1
    transition.target = location2
    location1.add_transition(transition)

    automaton = Automaton("automaton", frame, [location1, location2], location1)
    automaton_controller = AutomatonController(automaton)

    assert automaton_controller.step() == (None, None)
    assert chan.receiving_transitions[0] == transition
    assert len(automaton_controller._clean_up_list) == 1


def test_step_sending_binary_sync_has_pair():
    """
    Test step method with transition with sending binary transition with a pair.
    """
    frame = Frame()
    chan = Channel(False)
    frame.add_binding("x", chan)
    val = ID("x")
    synchronization = Synchronization(True, val)

    location1 = Location("id0", "location1")
    location2 = Location("id1", "location2")
    location3 = Location("id2", "location3")
    location4 = Location("id3", "location4")
    transition1 = Transition(synchronization = synchronization)
    transition1.source = location1
    transition1.target = location2
    location1.add_transition(transition1)
    transition2 = Transition(synchronization = synchronization)
    transition2.source = location3
    transition2.target = location4
    location3.add_transition(transition2)
    chan.register_receiving(transition2)

    automaton = Automaton("automaton", frame, [location1, location2], location1)
    automaton_controller = AutomatonController(automaton)

    assert automaton_controller.step() == (transition1, chan)
    assert len(automaton_controller._clean_up_list) == 1


def test_step_receiving_binary_sync_has_pair():
    """
    Test step method with transition with receiving binary transition with a pair.
    """
    frame = Frame()
    chan = Channel(False)
    frame.add_binding("x", chan)
    val = ID("x")
    synchronization = Synchronization(False, val)

    location1 = Location("id0", "location1")
    location2 = Location("id1", "location2")
    location3 = Location("id2", "location3")
    location4 = Location("id3", "location4")
    transition1 = Transition(synchronization = synchronization)
    transition1.source = location1
    transition1.target = location2
    location1.add_transition(transition1)
    transition2 = Transition(synchronization = synchronization)
    transition2.source = location3
    transition2.target = location4
    location3.add_transition(transition2)
    chan.register_sending(transition2)

    automaton = Automaton("automaton", frame, [location1, location2], location1)
    automaton_controller = AutomatonController(automaton)

    assert automaton_controller.step() == (transition1, chan)
    assert len(automaton_controller._clean_up_list) == 1


def test_clean_up_channels():
    """
    Test clean up channel method.
    """
    chan = Channel(False)
    location1 = Location("id0", "location1")
    location2 = Location("id0", "location2")
    location3 = Location("id2", "location3")
    transition1 = Transition()
    transition1.source = location1
    transition2 = Transition()
    transition2.source = location1
    transition3 = Transition()
    transition3.source = location2
    transition4 = Transition()
    transition4.source = location2
    transition1.target = location3
    transition2.target = location3
    transition3.target = location3
    transition4.target = location3
    location1.add_transition(transition1)
    location1.add_transition(transition2)
    location2.add_transition(transition3)
    location2.add_transition(transition4)

    chan.register_receiving(transition1)
    chan.register_receiving(transition3)
    chan.register_sending(transition2)
    chan.register_sending(transition4)

    frame = Frame()
    automaton = Automaton("automaton", frame, [location1], location1)
    automaton_controller = AutomatonController(automaton)
    automaton_controller._clean_up_list.append(chan)

    assert len(chan.sending_transitions) == 2
    assert len(chan.receiving_transitions) == 2

    automaton_controller.clean_up_channels()

    assert len(chan.sending_transitions) == 1
    assert len(chan.receiving_transitions) == 1


def test_find_controller():
    """
    Tests if the find_controller method works.
    """
    location = Location("id0", "location")
    transition = Transition()
    transition.source = location
    frame1 = Frame()
    frame2 = Frame()
    frame3 = Frame()
    automaton = Automaton("automaton", frame1, [location], location)
    system = System(frame2, frame3, [automaton])
    system_controller = SystemController(system)

    assert (system_controller._find_controller(transition) ==
            system_controller.automata_controller[0])


def test_execute_transition():
    """
    Test if the execute transition method changes the current location.
    """
    location1 = Location("id0", "location1")
    location2 = Location("id1", "location2")
    transition = Transition()
    transition.source = location1
    transition.target = location2
    location1.add_transition(transition)
    frame1 = Frame()
    frame2 = Frame()
    frame3 = Frame()
    automaton = Automaton("automaton", frame1, [location1, location2], location1)
    system = System(frame2, frame3, [automaton])
    system_controller = SystemController(system)

    assert system_controller.automata_controller[0].current_location == location1

    system_controller._execute_transition(transition, system_controller.automata_controller[0])

    assert system_controller.automata_controller[0].current_location == location2


def test_execute_binary_transition():
    """
    Test if the execute binary transition executes both transitions of the synchronization.
    """
    val = ID("x")
    synchronization1 = Synchronization(True, val)
    synchronization2 = Synchronization(False, val)
    location1 = Location("id0", "location1")
    location2 = Location("id1", "location2")
    location3 = Location("id2", "location3")
    location4 = Location("id3", "location4")
    transition1 = Transition(synchronization = synchronization1)
    transition1.source = location1
    transition1.target = location2
    location1.add_transition(transition1)
    transition2 = Transition(synchronization = synchronization2)
    transition2.source = location3
    transition2.target = location4
    location3.add_transition(transition2)
    chan = Channel(False)
    frame1 = Frame()
    frame1.add_binding("x", chan)
    frame2 = Frame()
    frame2.add_binding("x", chan)
    frame3 = Frame()
    frame4 = Frame()
    automaton1 = Automaton("automaton", frame1, [location1, location2], location1)
    automaton2 = Automaton("automaton", frame2, [location3, location4], location3)
    system = System(frame3, frame4, [automaton1, automaton2])
    system_controller = SystemController(system)

    step1 = system_controller.automata_controller[0].step()
    assert step1 == (None, None)
    step2 = system_controller.automata_controller[1].step()
    assert step2 == (transition2, chan)

    assert system_controller.automata_controller[0].current_location == location1
    assert system_controller.automata_controller[1].current_location == location3
    assert len(system_controller.automata_controller[0]._clean_up_list) == 1
    assert len(system_controller.automata_controller[1]._clean_up_list) == 1
    assert len(chan.sending_transitions) == 1
    assert len(chan.receiving_transitions) == 1

    system_controller._execute_binary_transition(transition2, chan,
                                                 system_controller.automata_controller[1])

    assert system_controller.automata_controller[0].current_location == location2
    assert system_controller.automata_controller[1].current_location == location4
    assert len(system_controller.automata_controller[0]._clean_up_list) == 0
    assert len(system_controller.automata_controller[1]._clean_up_list) == 0
    assert len(chan.receiving_transitions) == 0
    assert len(chan.sending_transitions) == 0


def test_execute_binary_transition_changed_channel():
    """
    Test if the execute binary transition still executes both transitions of the synchronization
    if channel reference changed before second transition was found.
    """
    val1 = ID("x")
    val2 = ArrayRef("a", val1)
    val3 = ID("y")
    synchronization1 = Synchronization(True, val2)
    synchronization2 = Synchronization(False, val3)
    location1 = Location("id0", "location1")
    location2 = Location("id1", "location2")
    location3 = Location("id2", "location3")
    location4 = Location("id3", "location4")
    transition1 = Transition(synchronization = synchronization1)
    transition1.source = location1
    transition1.target = location2
    location1.add_transition(transition1)
    transition2 = Transition(synchronization = synchronization2)
    transition2.source = location3
    transition2.target = location4
    location3.add_transition(transition2)
    chan1 = Channel(False)
    chan2 = Channel(False)
    array = Array(False, [chan1, chan2])
    integer = IntVariable(False, 0)
    frame1 = Frame()
    frame1.add_binding("x", integer)
    frame1.add_binding("a", array)
    frame2 = Frame()
    frame2.add_binding("y", chan1)
    frame3 = Frame()
    frame4 = Frame()
    automaton1 = Automaton("automaton", frame1, [location1, location2], location1)
    automaton2 = Automaton("automaton", frame2, [location3, location4], location3)
    system = System(frame3, frame4, [automaton1, automaton2])
    system_controller = SystemController(system)

    step1 = system_controller.automata_controller[0].step()
    assert step1 == (None, None)

    integer.value = 1

    step2 = system_controller.automata_controller[1].step()
    assert step2 == (transition2, chan1)

    assert system_controller.automata_controller[0].current_location == location1
    assert system_controller.automata_controller[1].current_location == location3
    assert len(system_controller.automata_controller[0]._clean_up_list) == 1
    assert len(system_controller.automata_controller[1]._clean_up_list) == 1
    assert len(chan1.sending_transitions) == 1
    assert len(chan1.receiving_transitions) == 1

    system_controller._execute_binary_transition(transition2, chan1,
                                                 system_controller.automata_controller[1])

    assert system_controller.automata_controller[0].current_location == location1
    assert system_controller.automata_controller[1].current_location == location3
    assert len(system_controller.automata_controller[0]._clean_up_list) == 1
    assert len(system_controller.automata_controller[1]._clean_up_list) == 1
    assert len(chan1.sending_transitions) == 1
    assert len(chan1.receiving_transitions) == 1


def test_execute_binary_transition_changed_can_step_result():
    """
    Test if the execute binary transition still executes both transitions of the synchronization
    if the result of can_step of the first transition has changed in the time between
    the first and second transition found.
    """
    val1 = ID("x")
    val2 = ID("y")
    val3 = Constant("int", 5)
    bin_op = BinaryOp(">", val2, val3)
    guard = Guard(bin_op)
    synchronization1 = Synchronization(True, val1)
    synchronization2 = Synchronization(False, val1)
    location1 = Location("id0", "location1")
    location2 = Location("id1", "location2")
    location3 = Location("id2", "location3")
    location4 = Location("id3", "location4")
    transition1 = Transition(guard = guard, synchronization = synchronization1)
    transition1.source = location1
    transition1.target = location2
    location1.add_transition(transition1)
    transition2 = Transition(synchronization = synchronization2)
    transition2.source = location3
    transition2.target = location4
    location3.add_transition(transition2)
    chan = Channel(False)
    integer = IntVariable(False, 10)
    frame1 = Frame()
    frame1.add_binding("x", chan)
    frame1.add_binding("y", integer)
    frame2 = Frame()
    frame2.add_binding("x", chan)
    frame3 = Frame()
    frame4 = Frame()
    automaton1 = Automaton("automaton", frame1, [location1, location2], location1)
    automaton2 = Automaton("automaton", frame2, [location3, location4], location3)
    system = System(frame3, frame4, [automaton1, automaton2])
    system_controller = SystemController(system)

    step1 = system_controller.automata_controller[0].step()
    assert step1 == (None, None)

    integer.value = 0

    step2 = system_controller.automata_controller[1].step()
    assert step2 == (transition2, chan)

    assert system_controller.automata_controller[0].current_location == location1
    assert system_controller.automata_controller[1].current_location == location3
    assert len(system_controller.automata_controller[0]._clean_up_list) == 1
    assert len(system_controller.automata_controller[1]._clean_up_list) == 1
    assert len(chan.sending_transitions) == 1
    assert len(chan.receiving_transitions) == 1

    system_controller._execute_binary_transition(transition2, chan,
                                                 system_controller.automata_controller[1])

    assert system_controller.automata_controller[0].current_location == location1
    assert system_controller.automata_controller[1].current_location == location3
    assert len(system_controller.automata_controller[0]._clean_up_list) == 1
    assert len(system_controller.automata_controller[1]._clean_up_list) == 1
    assert len(chan.sending_transitions) == 1
    assert len(chan.receiving_transitions) == 1


def test_execute_broadcast_transition():
    """
    Test the execution of the broadcast transition.
    """
    val = ID("x")
    synchronization1 = Synchronization(True, val)
    synchronization2 = Synchronization(False, val)
    location1 = Location("id0", "location1")
    location2 = Location("id1", "location2")
    location3 = Location("id2", "location3")
    location4 = Location("id3", "location4")
    transition1 = Transition(synchronization = synchronization1)
    transition1.source = location1
    transition1.target = location2
    location1.add_transition(transition1)
    transition2 = Transition(synchronization = synchronization2)
    transition2.source = location3
    transition2.target = location4
    location3.add_transition(transition2)
    chan = Channel(True)
    frame1 = Frame()
    frame1.add_binding("x", chan)
    frame2 = Frame()
    frame2.add_binding("x", chan)
    frame3 = Frame()
    frame4 = Frame()
    automaton1 = Automaton("automaton", frame1, [location1, location2], location1)
    automaton2 = Automaton("automaton", frame2, [location3, location4], location3)
    system = System(frame3, frame4, [automaton1, automaton2])
    system_controller = SystemController(system)

    step1 = system_controller.automata_controller[0].step()
    assert step1 == (transition1, chan)

    assert system_controller.automata_controller[0].current_location == location1
    assert system_controller.automata_controller[1].current_location == location3

    system_controller._execute_broadcast_transition(transition1, chan,
                                                    system_controller.automata_controller[0])

    assert system_controller.automata_controller[0].current_location == location2
    assert system_controller.automata_controller[1].current_location == location4
