"""
This module contains test cases for verifying the parsing and Abstract Syntax Tree (AST) generation
of UPPAAL declarations using the `DeclarationParser` and its related components.
Each test validates the specific structure and correctness of the produced AST
or various UPPAAL declaration scenarios.
"""

from uppaal2jetracer.parser.declarationparser.declaration_parser import DeclarationParser
from uppaal2jetracer.parser.declarationparser.plyparser import ParseError
from uppaal2jetracer.declarations.declarations_tree_visitor import DeclarationsTreeVisitor

parser = DeclarationParser(lex_optimize = False, yacc_optimize = False, yacc_debug = False)
tree_visitor = DeclarationsTreeVisitor()


# Basic declarations

def test_x_eq_y_plus_3():
    """
    Tests whether the given UPPAAL declaration code containing the declaration and assignment of
    variables `x` and `y` complies with the expected AST structure.
    """
    declaration_code = """
        int main() {
            int x = 0, y = 0;
            x = y + 3;
        }
    """
    declaration_ast = parser.parse(declaration_code, "", False)
    assert ("(FileAST: (FuncDef: (Decl: (FuncDecl: (TypeDecl: (IdentifierType: int) main)))("
            "(Compound: (Decl: (TypeDecl: (IdentifierType: int) x) = (Constant: 0)), "
            "(Decl: (TypeDecl: (IdentifierType: int) y) = (Constant: 0)), "
            "(Assignment: (ID: x) = ((BinaryOp: ((ID: y)) + ((Constant: 3))))))"
            ")))") == tree_visitor.visit_fileast(declaration_ast)


def test_x_eq_y_plus_3_no_function_fail():
    """
    Tests whether the given UPPAAL declaration code containing the declaration and assignment of
    variables `x` and `y` complies with the expected AST structure.
    In this test case, the `main` function is omitted.
    """
    declaration_code = """
            int x = 0, y = 0;
            x = y + 3;
        """
    try:
        parser.parse(declaration_code, "", False)
        assert False
    except ParseError:
        assert True


def test_add_function():
    """
    Tests whether the given UPPAAL declaration code containing the declaration and assignment of
    variables `x` and `y` complies with the expected AST structure.
    """
    declaration_code = """
        int add(int a, int b) {
            return a + b;
        }
    """
    declaration_ast = parser.parse(declaration_code, "", False)
    assert ("(FileAST: "
            "(FuncDef: (Decl: (FuncDecl: (TypeDecl: (IdentifierType: int) add)))("
            "(Compound: (Return: (BinaryOp: ((ID: a)) + ((ID: b)))))"
            "))"
            ")") == tree_visitor.visit_fileast(declaration_ast)


# Basic types

def test_const_int():
    """
    Tests the representation of a constant integer declaration in the AST.

    :raises AssertionError: If the parsed AST representation does not match the
        expected structure.
    """
    declaration_code = "const int myInteger = 42;"
    declaration_ast = parser.parse(declaration_code, "", False)
    assert ("(FileAST: "
            "(Decl: (TypeDecl: const (IdentifierType: int) myInteger) = (Constant: 42))"
            ")") == tree_visitor.visit_fileast(declaration_ast)


def test_string():
    """
    Tests parsing and visiting of string variable declaration using a parser and tree
    visitor. The function parses the given C-like code snippet defining a constant
    string variable, then verifies the corresponding abstract syntax tree (AST) is
    correctly generated by comparing it to the expected AST representation. The
    test ensures that the parser and visitor are working as intended for this kind
    of input.
    """
    declaration_code = "const string myString;"
    declaration_ast = parser.parse(declaration_code, "", False)
    assert ("(FileAST: "
            "(Decl: (TypeDecl: const (IdentifierType: string) myString))"
            ")") == tree_visitor.visit_fileast(declaration_ast)


def test_string_w_value():
    """
    Tests the parsing and AST (Abstract Syntax Tree) generation process of a
    C-like language's code snippet. Ensures the parsed syntax corresponds to
    the expected tree structure defined in the test case. The test checks
    whether the transformation from source code to its representation as
    an abstract syntax tree is functioning correctly.
    """
    declaration_code = 'const string myString = "42";'
    declaration_ast = parser.parse(declaration_code, "", False)
    assert ('(FileAST: '
            '(Decl: (TypeDecl: const (IdentifierType: string) myString) = (Constant: "42"))'
            ')') == tree_visitor.visit_fileast(declaration_ast)


def test_string_wo_const():
    """
    Tests the ability of the parser to correctly handle a string declaration
    without a constant value. This function verifies that the parser raises the
    appropriate `ParseError` exception when the input code does not conform to
    expected syntax and format.
    """
    declaration_code = "string myString;"
    try:
        parser.parse(declaration_code, "", False)
        assert False
    except ParseError:
        assert True


def test_struct():
    """
    Tests the parsing of a struct.

    This function verifies the correct interpretation and representation of a struct.
    It parses a given code snippet into an AST and compares its structure with
    the expected output to ensure accurate parsing and transformation by the system.

    :raises AssertionError: If the actual AST structure does not match the expected
                            structure.
    """
    declaration_code = ("struct A {"
                        "int i;"
                        "double d;"
                        "};")
    declaration_ast = parser.parse(declaration_code, "", False)
    assert ("(FileAST: "
            "(Decl: (Struct: A ("
            "(Decl: (TypeDecl: (IdentifierType: int) i))"
            "(Decl: (TypeDecl: (IdentifierType: double) d))"
            ")))"
            ")") == tree_visitor.visit_fileast(declaration_ast)


def test_typedef_struct():
    """
    Tests the parsing of a typedef with a struct.

    This function verifies the correct interpretation and representation of a typedef struct.
    It parses a given code snippet into an AST and compares its structure with
    the expected output to ensure accurate parsing and transformation by the system.

    :raises AssertionError: If the actual AST structure does not match the expected
                            structure.
    """
    declaration_code = ("typedef struct {"
                        "int x, y;"
                        "} Point;")
    declaration_ast = parser.parse(declaration_code, "", False)
    assert ("(FileAST: "
            "(Typedef: is_typedef=True (TypeDecl: "
            "(Struct: None ("
            "(Decl: (TypeDecl: (IdentifierType: int) x))"
            "(Decl: (TypeDecl: (IdentifierType: int) y))"
            ")) Point"
            "))"
            ")") == tree_visitor.visit_fileast(declaration_ast)


def test_typedef_type():
    """
    Tests the parsing of a typedef with a struct.

    This function verifies the correct interpretation and representation of a typedef struct.
    It parses a given code snippet into an AST and compares its structure with
    the expected output to ensure accurate parsing and transformation by the system.

    :raises AssertionError: If the actual AST structure does not match the expected
                            structure.
    """
    declaration_code = ("typedef int myint;"
                        "myint x = 10;")
    declaration_ast = parser.parse(declaration_code, "", False)
    assert ("(FileAST: "
            "(Typedef: is_typedef=True "
            "(TypeDecl: (IdentifierType: int) myint)), "
            "(Decl: (TypeDecl: (IdentifierType: myint) x) = (Constant: 10)"
            ")"
            ")") == tree_visitor.visit_fileast(declaration_ast)


def test_boolean_true():
    """
    Validates the parsing of a boolean true value.

    This function tests whether a parsed boolean variable declaration with the value
    `true` is correctly converted into an abstract syntax tree (AST). The function
    constructs a boolean variable declaration as code, parses it with a parser, and
    compares the resulting AST with the expected string representation. The success
    of the test is determined by asserting that the actual and expected outputs
    match.

    :raises AssertionError: If the parsed AST does not match the expected
        structure.
    """
    declaration_code = "bool b = true;"
    declaration_ast = parser.parse(declaration_code, "", False)
    assert ("(FileAST: "
            "(Decl: (TypeDecl: (IdentifierType: bool) b) = (Constant: true))"
            ")") == tree_visitor.visit_fileast(declaration_ast)


def test_boolean_false():
    """
    Tests that a boolean variable declaration with the value `false` is parsed
    and visits correctly within the AST context. This involves parsing a code
    string, producing the AST, and verifying the structure of the parsed and
    visited outcome against the expected output.

    This test ensures the parser can correctly distinguish and process boolean
    declarations with `false` assignments, generating the accurate tree structure
    during the visit phase.
    """
    declaration_code = "bool b = false;"
    declaration_ast = parser.parse(declaration_code, "", False)
    assert ("(FileAST: "
            "(Decl: (TypeDecl: (IdentifierType: bool) b) = (Constant: false))"
            ")") == tree_visitor.visit_fileast(declaration_ast)


def test_boolean_w_nonbool_value_fail():
    """
    Tests that parsing a boolean declaration with a non-boolean value fails as expected.

    This test ensures that the parser raises a `ParseError` exception when attempting to
    parse a boolean variable initialized with a non-boolean value. Specifically, the test
    attempts to declare a boolean variable `b` and assign it a non-boolean integer value
    42.

    :raises ParseError: when the parser encounters an invalid boolean assignment.
    """
    declaration_code = "bool b = 42;"
    try:
        parser.parse(declaration_code, "", False)
        assert False
    except ParseError:
        assert True


def test_nonbool_w_boolean_value_fail():
    """
    Tests the scenario where a non-boolean variable is incorrectly assigned
    a boolean value. The test ensures that the parser raises a `ParseError`
    when such a case is encountered.

    The test defines a declaration statement with an erroneous type assignment,
    attempts to parse it, and subsequently verifies that the expected error is
    thrown by the parser.

    :raises ParseError: if the parser correctly identifies the assignment
        of a boolean value to a non-boolean type and raises a parsing error.
    """
    declaration_code = "int a = true;"
    try:
        parser.parse(declaration_code, "", False)
        assert False
    except ParseError:
        assert True


# UPPAAL types

def test_channel():
    """
    Tests the parsing and visitation of a channel declaration AST (Abstract Syntax Tree).

    This function validates the parsing process of a channel declaration string and
    checks the resulting AST structure against the expected output.

    :raises AssertionError: If the resulting parsed AST structure does not match the
        expected format.
    """
    declaration_code = "chan myChannel;"
    declaration_ast = parser.parse(declaration_code, "", False)
    assert ("(FileAST: "
            "(Decl: (TypeDecl: (IdentifierType: chan) myChannel))"
            ")") == tree_visitor.visit_fileast(declaration_ast)


def test_channel_w_value_fail():
    """
    Tests the failure scenario for channel declaration with a value.

    :raises ParseError: If the parser correctly identifies a channel
        declaration with an invalid value as erroneous.
    """
    declaration_code = "chan myChannel = 42;"
    try:
        parser.parse(declaration_code, "", False)
        assert False
    except ParseError:
        assert True


def test_channel_broadcast():
    """
    Tests the channel broadcasting declaration.

    This function is specifically designed to validate the correctness of the
    channel broadcasting syntax parsing and AST creation.
    """
    declaration_code = "broadcast chan myChannel;"
    declaration_ast = parser.parse(declaration_code, "", False)
    assert ("(FileAST: "
            "(Decl: (TypeDecl: broadcast (IdentifierType: chan) myChannel))"
            ")") == tree_visitor.visit_fileast(declaration_ast)


def test_qualifier_broadcast_w_nonchan_type_fail():
    """
    Tests that broadcasting a non-channel type using a broadcast qualifier fails.

    This function attempts to parse a declaration code with a broadcast qualifier
    applied to a non-channel type. If the parsing succeeds, the test explicitly
    fails by asserting False. When a `ParseError` is raised due to invalid usage
    of the broadcast modifier, the test passes. The test ensures that such invalid
    declarations are properly caught and flagged by the parser.

    :raises ParseError: If the parsing fails due to invalid use of the broadcast
        qualifier, the error is raised as expected.
    """
    declaration_code = "broadcast int myChannel;"
    try:
        parser.parse(declaration_code, "", False)
        assert False
    except ParseError:
        assert True


def test_clock():
    """
    Tests the functionality of the C code parser and the tree visitor by parsing a
    declaration code snippet and asserting that the generated abstract syntax tree
    (AST) matches the expected AST structure.

    This test ensures that the `parser` object correctly recognizes C declarations
    and that the resulting AST representation aligns with the expected tree format,
    validating both parsing and tree visiting mechanisms.

    :raises AssertionError: If the generated AST does not match the expected AST
        structure.
    """
    declaration_code = "clock myClock;"
    declaration_ast = parser.parse(declaration_code, "", False)
    assert ("(FileAST: "
            "(Decl: (TypeDecl: (IdentifierType: clock) myClock))"
            ")") == tree_visitor.visit_fileast(declaration_ast)


def test_clock_w_nonint_value_fail():
    """
    Tests parsing of a clock declaration when provided with a non-integer value.
    This test ensures that the parser raises the appropriate error when the
    clock is initialized incorrectly with a string instead of an integer.
    """
    declaration_code = "clock myClock = '42';"
    try:
        parser.parse(declaration_code, "", False)
        assert False
    except ParseError:
        assert True


def test_range_2_3():
    """
    Tests the parsing, AST generation, and AST visitation for a code snippet declaring a
    ranged integer array.

    This test case ensures that the parser correctly interprets a declaration involving a ranged
    integer array in the format `int[2, 3]`. The subsequent AST representation of the code snippet
    is compared against the expected structure to verify the correctness of the parsing process
    and the AST visitor.

    :raises AssertionError: If the actual AST produced by the parser and visitor differs from the
        expected AST defined in the test case.
    """
    declaration_code = "int[2, 3] a;"
    declaration_ast = parser.parse(declaration_code, "", False)
    assert ("(FileAST: "
            "(Decl: (RangeDecl: "
            "(TypeDecl: (IdentifierType: int) a) [(Constant: 2), (Constant: 3)]))"
            ")") == tree_visitor.visit_fileast(declaration_ast)


def test_range_2_3_w_spaces_fail():
    """
    Tests the parsing of a declaration code that specifies a variable of type 'int'
    with a range '[2, 3]' that includes spaces. The function verifies that the parser
    correctly raises a `ParseError` for invalid syntax, ensuring robust handling of
    such input.

    :raises ParseError: Raised when the parser encounters invalid syntax.
    """
    declaration_code = "int [2, 3] a;"
    try:
        parser.parse(declaration_code, "", False)
        assert False
    except ParseError:
        assert True


def test_range_2_3_double_type_fail():
    """
    Tests the failure of parsing a double type declaration with an invalid range.

    This test ensures that the parser raises a `ParseError` when attempting to parse
    a declaration for a floating-point type with a range `[2, 3]`, which does not
    comply with expected formatting or constraints.
    """
    declaration_code = "double[2, 3] a;"
    try:
        parser.parse(declaration_code, "", False)
        assert False
    except ParseError:
        assert True


def test_range_3_2():
    """
    Test the parsing and AST generation for a ranged declaration and validate
    its visitation using the tree visitor.

    The function validates proper parsing of a declaration code string with a
    range type, generating its abstract syntax tree (AST), and ensuring that
    the tree visitor produces a correct and expected representation of the
    parsed AST.

    :raises AssertionError: If the visitation result does not match the expected
                             representation of the ranged declaration's AST.
    """
    declaration_code = "int[3, 2] a;"
    declaration_ast = parser.parse(declaration_code, "", False)
    assert ("(FileAST: "
            "(Decl: (RangeDecl: "
            "(TypeDecl: (IdentifierType: int) a) [(Constant: 3), (Constant: 2)]))"
            ")") == tree_visitor.visit_fileast(declaration_ast)


def test_range_0_N():
    """
    Executes a test to verify the parsing and AST transformation of a declaration
    code defining a range of integers from 0 to N. The test parses the input code
    and compares the resulting Abstract Syntax Tree (AST) with the expected output.

    The purpose of this test is to ensure that the parser and visitor accurately
    handle range declarations and produce the correct AST structure.

    :raises AssertionError: If the parsed AST does not match the expected AST.
    """
    declaration_code = "int[0, N] a;"
    declaration_ast = parser.parse(declaration_code, "", False)
    assert ("(FileAST: "
            "(Decl: (RangeDecl: "
            "(TypeDecl: (IdentifierType: int) a) [(Constant: 0), (ID: N)]))"
            ")") == tree_visitor.visit_fileast(declaration_ast)


def test_range_N_0():
    """
    Executes a test to verify the parsing and AST transformation of a declaration
    code defining a range of integers from N to 0. The test parses the input code
    and compares the resulting Abstract Syntax Tree (AST) with the expected output.

    The purpose of this test is to ensure that the parser and visitor accurately
    handle range declarations and produce the correct AST structure.

    :raises AssertionError: If the parsed AST does not match the expected AST.
    """
    declaration_code = "int[N, 0] a;"
    declaration_ast = parser.parse(declaration_code, "", False)
    assert ("(FileAST: "
            "(Decl: (RangeDecl: "
            "(TypeDecl: (IdentifierType: int) a) [(ID: N), (Constant: 0)]))"
            ")") == tree_visitor.visit_fileast(declaration_ast)


def test_range_N_N():
    """
    Executes a test to verify the parsing and AST transformation of a declaration
    code defining a range of integers from N to N. The test parses the input code
    and compares the resulting Abstract Syntax Tree (AST) with the expected output.

    The purpose of this test is to ensure that the parser and visitor accurately
    handle range declarations and produce the correct AST structure.

    :raises AssertionError: If the parsed AST does not match the expected AST.
    """
    declaration_code = "int[N, N] a;"
    declaration_ast = parser.parse(declaration_code, "", False)
    assert ("(FileAST: "
            "(Decl: (RangeDecl: "
            "(TypeDecl: (IdentifierType: int) a) [(ID: N), (ID: N)]))"
            ")") == tree_visitor.visit_fileast(declaration_ast)


def test_range_N_M():
    """
    Executes a test to verify the parsing and AST transformation of a declaration
    code defining a range of integers from N to M. The test parses the input code
    and compares the resulting Abstract Syntax Tree (AST) with the expected output.

    The purpose of this test is to ensure that the parser and visitor accurately
    handle range declarations and produce the correct AST structure.

    :raises AssertionError: If the parsed AST does not match the expected AST.
    """
    declaration_code = "int[N, M] a;"
    declaration_ast = parser.parse(declaration_code, "", False)
    assert ("(FileAST: "
            "(Decl: (RangeDecl: "
            "(TypeDecl: (IdentifierType: int) a) [(ID: N), (ID: M)]))"
            ")") == tree_visitor.visit_fileast(declaration_ast)


def test_range_qualifier():
    """
    Tests the range qualifier for a declaration in a parsed abstract syntax tree (AST).

    This function verifies the correct interpretation and representation of a range
    qualifier in a C-style declaration. It parses a given code snippet into an AST
    and compares its structure with the expected output to ensure accurate parsing
    and transformation by the system.

    :raises AssertionError: If the actual AST structure does not match the expected
                            structure.
    """
    declaration_code = "const int[2, 3] a;"
    declaration_ast = parser.parse(declaration_code, "", False)
    assert ("(FileAST: "
            "(Decl: (RangeDecl: "
            "(TypeDecl: const (IdentifierType: int) a) [(Constant: 2), (Constant: 3)]))"
            ")") == tree_visitor.visit_fileast(declaration_ast)


# Negative features

def test_cast_fail():
    """
    Tests if casting of variable types fails as it's not supported by UPPAAL.

    :raises ParseError: If the parser correctly parses a variable cast.
    """
    declaration_code = "double mean = (double) sum / count;"
    try:
        parser.parse(declaration_code, "", False)
        assert False
    except ParseError:
        assert True


def test_ellipsis_param_fail():
    """
    Tests if ellipsis as parameter specification fails as it's not supported by UPPAAL.

    :raises ParseError: If the parser correctly parses an ellipsis.
    """
    declaration_code = "void format(const char* fmt, ...);"
    try:
        parser.parse(declaration_code, "", False)
        assert False
    except ParseError:
        assert True


def test_enum_fail():
    """
    Tests if parsing of enums fails as they are not supported by UPPAAL.

    :raises ParseError: If the parser correctly parses an enum.
    """
    declaration_code = "enum Level { LOW, MEDIUM, HIGH };"
    try:
        parser.parse(declaration_code, "", False)
        assert False
    except ParseError:
        assert True


def test_goto_fail():
    """
    Tests if the parsing of a goto statement fails as they are not supported by UPPAAL.

    :raises ParseError: If the parser correctly parses a goto statement.
    """
    declaration_code = "goto stop;"
    try:
        parser.parse(declaration_code, "", False)
        assert False
    except ParseError:
        assert True


def test_label_fail():
    """
    Tests if parsing of labels fails as they are not supported by UPPAAL.

    :raises ParseError: If the parser correctly parses a label.
    """
    declaration_code = "stop: add(5, 6);"
    try:
        parser.parse(declaration_code, "", False)
        assert False
    except ParseError:
        assert True


def test_ptr_decl_fail():
    """
    Tests if parsing of pointers fails as pointers are not supported by UPPAAL.

    :raises ParseError: If the parser correctly parses a pointer declaration.
    """
    declaration_code = "int* p = pointer;"
    try:
        parser.parse(declaration_code, "", False)
        assert False
    except ParseError:
        assert True


def test_static_assert_fail():
    """
    Tests if parsing of static asserts fails as they are not supported by UPPAAL.

    :raises ParseError: If the parser correctly parses a static assert.
    """
    declaration_code = '_Static_assert(2 + 2 * 2 == 6, "Hello World");'
    try:
        parser.parse(declaration_code, "", False)
        assert False
    except ParseError:
        assert True


def test_switch_case_default_fail():
    """
    Tests if parsing of a switch case default fails as they are not supported by uppaal2jetracer.

    :raises ParseError: If the parser correctly parses a switch case default.
    """
    declaration_code = ("switch(c) {"
                        "case 'A':  "
                        "   capital_a++;"
                        "default:"
                        "   total++;"
                        "}")
    try:
        parser.parse(declaration_code, "", False)
        assert False
    except ParseError:
        assert True


def test_break_fail():
    """
    Tests if parsing of a break statement fails as they are not supported by UPPAAL.

    :raises ParseError: If the parser correctly parses a break statement.
    """
    declaration_code = "break;"
    try:
        parser.parse(declaration_code, "", False)
        assert False
    except ParseError:
        assert True


def test_continue_fail():
    """
    Tests if parsing of a continue statement fails as they are not supported by UPPAAL.

    :raises ParseError: If the parser correctly parses a continue statement.
    """
    declaration_code = "continue;"
    try:
        parser.parse(declaration_code, "", False)
        assert False
    except ParseError:
        assert True


def test_ternaryop_fail():
    """
    Tests if parsing of ternary operations fails as they are not supported by UPPAAL.

    :raises ParseError: If the parser correctly parses a ternary operation.
    """
    declaration_code = "c = (a < b) ? a : b;"
    try:
        parser.parse(declaration_code, "", False)
        assert False
    except ParseError:
        assert True


def test_union_fail():
    """
    Tests if parsing of unions fails as they are not supported by UPPAAL.

    :raises ParseError: If the parser correctly parses a union.
    """
    declaration_code = ("union A {"
                        "int i;"
                        "double d;"
                        "};")
    try:
        parser.parse(declaration_code, "", False)
        assert False
    except ParseError:
        assert True

def test_typedef_function_fail():
    """
    Tests if parsing of typedef function fails.

    :raises ParseError: If the parser correctly parses a function declared typedef.
    """
    declaration_code = "typedef int bar() { return 42; }"
    try:
        parser.parse(declaration_code, "", False)
        assert False
    except ParseError:
        assert True


def test_pragma_fail():
    """
    Tests if parsing of preprocessor pragma statements fails as they are not supported by UPPAAL.

    :raises ParseError: If the parser correctly parses a preprocessor pragma statement.
    """
    declaration_code = "#pragma once"
    try:
        parser.parse(declaration_code, "", False)
        assert False
    except ParseError:
        assert True

def test_postincrement_fail():
    """
    Tests if parsing of postincrement statements fails as they are not supported by UPPAAL.

    :raises ParseError: If the parser correctly parses a postincrement statement.
    """
    declaration_code = "x++;"
    try:
        parser.parse(declaration_code, "", False)
        assert False
    except ParseError:
        assert True


def test_preincrement_fail():
    """
    Tests if parsing of preincrement statements fails as they are not supported by UPPAAL.

    :raises ParseError: If the parser correctly parses a preincrement statement.
    """
    declaration_code = "++x;"
    try:
        parser.parse(declaration_code, "", False)
        assert False
    except ParseError:
        assert True


def test_postdecrement_fail():
    """
    Tests if parsing of postdecrement statements fails as they are not supported by UPPAAL.

    :raises ParseError: If the parser correctly parses a postdecrement statement.
    """
    declaration_code = "x--;"
    try:
        parser.parse(declaration_code, "", False)
        assert False
    except ParseError:
        assert True


def test_predecrement_fail():
    """
    Tests if parsing of predecrement statements fails as they are not supported by UPPAAL.

    :raises ParseError: If the parser correctly parses a predecrement statement.
    """
    declaration_code = "--x;"
    try:
        parser.parse(declaration_code, "", False)
        assert False
    except ParseError:
        assert True
