"""TransitionParser test

This file provides tests for a TransitionParser.
"""

from bs4 import BeautifulSoup

from uppaal2jetracer.declarations.declarations_tree_visitor import DeclarationsTreeVisitor
from uppaal2jetracer.parser.parser import TransitionParser
from uppaal2jetracer.uppaalmodel.elements import Transition, Synchronization, Guard, Update

parser = TransitionParser()
tree_visitor: DeclarationsTreeVisitor = DeclarationsTreeVisitor()


def test_parse_transition():
    """
    Test that the transition parser parses a transition as expected.
    """
    transition_tag = """\t\t<transition>
\t\t\t<source ref="id2"/>
\t\t\t<target ref="id1"/>
\t\t\t<label kind="synchronisation" x="-224" y="-248">shut_down?</label>
\t\t</transition>
        """
    transition = parser.parse(BeautifulSoup(transition_tag, "lxml-xml").find("transition"))
    assert isinstance(transition, Transition)


def test_parse_sync():
    """
    Test that the transition parser parses synchronizations as expected.
    """

    transition_tag = """\t\t<transition>
\t\t\t<source ref="id2"/>
\t\t\t<target ref="id1"/>
\t\t\t<label kind="synchronisation" x="-224" y="-248">shut_down?</label>
\t\t</transition>
    """
    transition = parser.parse(BeautifulSoup(transition_tag, "lxml-xml").find("transition"))
    assert isinstance(transition, Transition)
    assert isinstance(transition.synchronization, Synchronization)
    assert ("(FileAST: "
            "(ID: shut_down)"
            ")") == transition.synchronization.synchronization.accept(tree_visitor)
    assert transition.update is None
    assert transition.guard is None


def test_parse_guard():
    """
    Test that the transition parser parses guards as expected.
    """

    transition_tag = """\t\t<transition>
\t\t\t<source ref="id2"/>
\t\t\t<target ref="id1"/>
\t\t\t<label kind="guard" x="104" y="24">!set</label>
\t\t\t<label kind="assignment" x="104" y="48">x = 0, y = 0</label>
\t\t</transition>"""

    transition = parser.parse(BeautifulSoup(transition_tag, "lxml-xml").find("transition"))
    assert isinstance(transition, Transition)
    assert transition.synchronization is None
    assert isinstance(transition.guard, Guard)
    assert ("(FileAST: "
            "(UnaryOp: !(ID: set))"
            ")") == transition.guard.decision.accept(tree_visitor)
    assert isinstance(transition.update, Update)
    assert len(transition.update.assignments) == 2
    assert ("(FileAST: "
            "(Assignment: (ID: x) = ((Constant: 0)))"
            ")") == transition.update.assignments[0].accept(tree_visitor)
    assert ("(FileAST: "
            "(Assignment: (ID: y) = ((Constant: 0)))"
            ")") == transition.update.assignments[1].accept(tree_visitor)


def test_parse_update():
    """
    Test that the transition parser parses updates as expected.
    """

    transition_tag = """\t\t<transition>
\t\t\t<source ref="id1"/>
\t\t\t<target ref="id0"/>
\t\t\t<label kind="synchronisation" x="-224" y="-248">shut_down?</label>
\t\t\t<label kind="guard" x="144" y="72">x&lt;=k</label>
\t\t\t<label kind="assignment" x="144" y="88">x = 0, id = pid, set = true</label>
\t\t</transition>
    """

    transition = parser.parse(BeautifulSoup(transition_tag, "lxml-xml").find("transition"))
    assert isinstance(transition, Transition)
    assert isinstance(transition.synchronization, Synchronization)
    assert ("(FileAST: "
            "(ID: shut_down)"
            ")") == transition.synchronization.synchronization.accept(tree_visitor)
    assert isinstance(transition.guard, Guard)
    assert ("(FileAST: "
            "(BinaryOp: ((ID: x)) <= ((ID: k)))"
            ")") == transition.guard.decision.accept(tree_visitor)
    assert isinstance(transition.update, Update)
    assert len(transition.update.assignments) == 3
    assert ("(FileAST: "
            "(Assignment: (ID: x) = ((Constant: 0)))"
            ")") == transition.update.assignments[0].accept(tree_visitor)
    assert ("(FileAST: "
            "(Assignment: (ID: id) = ((ID: pid)))"
            ")") == transition.update.assignments[1].accept(tree_visitor)
    assert ("(FileAST: "
            "(Assignment: (ID: set) = ((Constant: true)))"
            ")") == transition.update.assignments[2].accept(tree_visitor)
