"""Version Manager test

This file provides tests for the version manager.
"""

import os
from typing import Tuple
from sqlalchemy import select

from uppaal2jetracer.versioncontrol.database import DatabaseConnection
from uppaal2jetracer.versioncontrol.versioncontrol import VersionManager, ProjectManager, \
    ResponseType
from uppaal2jetracer.versioncontrol.models import Version

TEST_DATABASE_RELATIVE_PATH = "data" + os.sep + "database"
TEST_DATABASE_NAME = "test_version"
TEST_DATABASE_FILE_TYPE = ".db"
TEST_DATABASE_SEPERATOR = "\\"


class FilePlaceHolder:
    """
    Just a Place-Holder for the later parsed python Object before serialization.
    """

    def __init__(self):
        pass


def create_database() -> Tuple[ProjectManager, VersionManager]:
    """
    Create the test database.
    """

    db = DatabaseConnection(TEST_DATABASE_RELATIVE_PATH, TEST_DATABASE_NAME)
    pm = ProjectManager(db)
    vm = VersionManager(db)

    return (pm, vm)


def remove_database(vm: VersionManager):
    """
    Remove the test database.
    """

    vm.db.session.close()
    vm.db.engine.dispose()

    remove_database_file(vm)


def remove_database_file(vm: VersionManager):
    """
    Remove the test database file.
    """

    vm.db.remove_db()


def test_add_version():
    """
    Test for adding a version.
    """

    tools = create_database()
    pm = tools[0]
    vm = tools[1]

    version_name = "My_Version"
    file_placeholder = FilePlaceHolder()

    no_project_version = vm.add_version(version_name, file_placeholder)

    assert no_project_version[0].type == ResponseType.ERROR

    project_name = "Test_Project"

    pm.add_project(project_name)

    valid_version = vm.add_version(version_name, file_placeholder)

    assert valid_version[0].type == ResponseType.OK
    assert valid_version[0].payload.name == version_name

    select_version = select(Version).where(Version._name == version_name)
    version = vm.db.session.scalars(select_version).first()

    assert version is not None and version.name == version_name

    test_version_limit = 2

    pm.change_version_limit(test_version_limit)

    version_name_template = "My_Version_"

    for i in range(1, test_version_limit):
        version_name_temp = version_name_template + str(i)
        version_data = vm.add_version(version_name_temp, file_placeholder)

        vm.favorite_version(version_data[0].payload.v_id)

    test_replace_version = "Replace_Version"

    replace_version = vm.add_version(test_replace_version, file_placeholder)

    assert replace_version[0].type == ResponseType.OK
    assert replace_version[0].payload.name == test_replace_version
    assert (replace_version[1].payload is not None
            and replace_version[1].payload.name == version_name)

    too_many_versions_version = "Too_Many_Version"

    vm.favorite_version(replace_version[0].payload.v_id)

    too_many_version = vm.add_version(too_many_versions_version, file_placeholder)

    assert too_many_version[0].type == ResponseType.ERROR

    invalid_version_name = "=inv;alid :version:name"

    invalid_name_version = vm.add_version(invalid_version_name, file_placeholder)

    assert invalid_name_version[0].type == ResponseType.ERROR

    remove_database(vm)


def test_get_all_versions():
    """
    Test for getting all versions.
    """

    tools = create_database()
    pm = tools[0]
    vm = tools[1]

    no_current_versions = vm.get_all_versions()

    assert no_current_versions[0].type == ResponseType.ERROR

    project_name = "Test_Project"

    new_project = pm.add_project(project_name)

    no_versions = vm.get_all_versions()

    assert len(no_versions) == 0

    num_versions = 3

    version_name_template = "My_Version_"
    file_placeholder = FilePlaceHolder()

    version_names = []

    for i in range(1, num_versions + 1):
        version_name = version_name_template + str(i)
        vm.add_version(version_name, file_placeholder)
        version_names.append(version_name)

    all_versions = vm.get_all_versions()

    assert len(all_versions) == num_versions

    for i in range(1, num_versions + 1):
        assert all_versions[num_versions - i].type == ResponseType.OK
        assert all_versions[num_versions - i].payload.name == version_names[i - 1]

    new_project_name = "Test_Project_New"

    pm.add_project(new_project_name)

    new_all_versions = vm.get_all_versions(new_project.payload.p_id)

    assert len(new_all_versions) == num_versions

    for i in range(1, num_versions + 1):
        assert all_versions[num_versions - i].type == ResponseType.OK
        assert all_versions[num_versions - i].payload.name == version_names[i - 1]

    invalid_id = -1

    invalid_versions = vm.get_all_versions(invalid_id)

    assert invalid_versions[0].type == ResponseType.ERROR

    remove_database(vm)


def test_get_version():
    """
    Test for getting a version.
    """

    tools = create_database()
    pm = tools[0]
    vm = tools[1]

    project_name = "Test_Project"

    pm.add_project(project_name)

    version_name = "My_Version"
    file_placeholder = FilePlaceHolder()

    new_version = vm.add_version(version_name, file_placeholder)

    found_version = vm.get_version(new_version[0].payload.v_id)

    assert found_version.type == ResponseType.OK
    assert found_version.payload.name == version_name

    invalid_id = -1

    invalid_version = vm.get_version(invalid_id)

    assert invalid_version.type == ResponseType.ERROR

    remove_database(vm)


def test_delete_version():
    """
    Test for deleting a version.
    """

    tools = create_database()
    pm = tools[0]
    vm = tools[1]

    project_name = "Test_Project"

    pm.add_project(project_name)

    version_name = "My_Version"
    file_placeholder = FilePlaceHolder()

    new_version = vm.add_version(version_name, file_placeholder)

    deleted_version = vm.delete_version(new_version[0].payload.v_id)

    assert deleted_version.type == ResponseType.OK
    assert deleted_version.payload.name == version_name

    select_version = select(Version).where(Version._name == version_name)
    version = vm.db.session.scalars(select_version).first()

    assert version is None

    invalid_id = -1

    invalid_version = vm.get_version(invalid_id)

    assert invalid_version.type == ResponseType.ERROR

    remove_database(vm)


def test_favorite_version():
    """
    Test for favoriting a version.
    """

    tools = create_database()
    pm = tools[0]
    vm = tools[1]

    project_name = "Test_Project"

    pm.add_project(project_name)

    version_name = "My_Version"
    file_placeholder = FilePlaceHolder()

    new_version = vm.add_version(version_name, file_placeholder)

    fav_version = vm.favorite_version(new_version[0].payload.v_id)

    assert fav_version.type == ResponseType.OK
    assert fav_version.payload.favorite is True

    fav_version_new = vm.favorite_version(new_version[0].payload.v_id)

    assert fav_version_new.type == ResponseType.OK
    assert fav_version_new.payload.favorite is False

    invalid_id = -1

    invalid_version = vm.get_version(invalid_id)

    assert invalid_version.type == ResponseType.ERROR

    remove_database(vm)
