"""Routes

This file holds all HTTP endpoints used by the user.
"""

import re
import os

from urllib.parse import urlsplit
from typing import Callable, Tuple
from functools import wraps
from threading import Thread

import sqlalchemy as sa
from flask import (render_template, send_file, send_from_directory, flash, redirect,
                   url_for, request, make_response, jsonify, Response)
from flask_login import current_user, login_user, logout_user, login_required
from werkzeug.utils import secure_filename

from uppaal2jetracer.command_system import CommandResultType
from uppaal2jetracer.controller.executor import Executor
from uppaal2jetracer.versioncontrol.versioncontrol import ResponseType

from uppaal2jetracer.webinterface.app.models import Global
from uppaal2jetracer.webinterface.app.forms import LoginForm
from uppaal2jetracer.webinterface.app.sockets import update_socket
from uppaal2jetracer.webinterface.app import app, db, webhandler

basedir = os.path.abspath(os.path.dirname(__file__))


HANDBUCH_RELATIVE_PATH = "assets" + os.sep + "pdf" + os.sep
HANDBUCH_FILE_NAME = "u2j_manual.pdf"

ICON_RELATIVE_PATH = "static" + os.sep + "icons"
ICON_FILE_NAME = "logo_basic.ico"

ERROR_TAG = "error"
TYPE_TAG = "type"
PAYLOAD_TAG = "payload"
MESSAGE_TAG = "message"

ERROR_TEMPLATE_MSG = "Error: {}"
UNAUTHORIZED_MSG = "Unauthorized."
INVALID_FILE_NAME_FORMAT_MSG = "Invalid file name."
INVALID_FILE_EXTENSION_MSG = "Invalid file extension."

PY_TRUE_MSG = "True"

UNAUTHORIZED_CODE = 401
COMMAND_ERROR_CODE = 400
SUCCESS_CODE = 200

LOGIN_ENDPOINT = "login"
MAIN_ENDPOINT = "index"

OPEN_PARENTHESIS = "("
CLOSE_PARENTHESIS = ")"

GET_ALL_ARG = "list"
RUN_ARG = "run"
FAVORITE_ARG = "fav"
DELETE_ARG = "delete"
DELETE_SKIP_CONF_ARG = "-s"
OPEN_ARG = "open"
NEW_ARG = "new"
LIMIT_ARG = "max"
CONFIG_ARG = "config"
CURRENT_ARG = "current"
STOP_ARG = "stop"


def get_file_number_indices(filename: str) -> Tuple[int]:
    """
    Gets the indices of the parenthesis of a file name if numbered.

    :param filename: Name of the file.
    :type filename: str
    :return: Pair of the indices of the surrounding parenthesis.
    :rtype: Tuple[int]
    """

    end_index = len(filename) - 1
    default_tup = (end_index, end_index)

    if filename[end_index] != CLOSE_PARENTHESIS or len(filename) == 1:
        return default_tup

    for i in range(2, len(filename) + 1):

        _index = len(filename) - i

        if filename[_index] == OPEN_PARENTHESIS:
            return (_index, end_index)

        if not filename[_index].isdigit():
            return default_tup

    return default_tup


def increase_file_number(filename: str, file_path: str, file_extension: str) -> str:
    """
    Increases the number behind a file name if the file already exists to negate conflicts.

    :param filename: Name of the file.
    :type filename: str
    :param file_path: Path of the file.
    :type file_path: str
    :param file_extension: Extension of the file.
    :type file_extension: str
    :return: The newly numbered file name.
    :rtype: str
    """

    indices = get_file_number_indices(filename)

    num = 1

    if indices[1] - indices[0] != 0:
        num = int(filename[indices[0] + 1: indices[1]])

    new_file_name = None

    while True:
        new_file_name = filename[:indices[0] + 1] + OPEN_PARENTHESIS + str(num) + CLOSE_PARENTHESIS

        new_file_path = os.path.join(file_path, new_file_name + file_extension)

        if not os.path.isfile(new_file_path):
            break

        num = num + 1

    return new_file_name


def string_to_bool(val: str) -> bool:
    """
    Parse a string to a boolean.

    :param val: The string to parse.
    :type val: str
    :return: Whether the parsed string as a boolean.
    :rtype: bool
    """

    return val == PY_TRUE_MSG


def create_payload_response(payload: dict, message: str, code: int = SUCCESS_CODE) -> Response:
    """
    Create a positive response with a given payload and response code.

    :param payload: The payload to return.
    :type payload: dict
    :param message: The success message to return.
    :type message: str
    :param code: The HTTP response code, defaults to 200
    :type code: int, optional
    :return: The HTTP response with the payload and code.
    :rtype: Response
    """

    return jsonify({PAYLOAD_TAG: payload, MESSAGE_TAG: message, TYPE_TAG: True}), code

def create_error_response(message: str, code: int) -> Response:
    """
    Create a negative response with a given message and response code.

    :param message: The error message to return.
    :type message: str
    :param code: The HTTP error response code
    :type code: int
    :return: The HTTP response with the error message and code.
    :rtype: Response
    """

    return jsonify({ ERROR_TAG: ERROR_TEMPLATE_MSG.format(message), TYPE_TAG: False }), code



def check_auth_token(f: Callable) -> Response:
    """
    Check whether the authorization header has the correct key.

    :param f: The endpoint function to protect with an auth key.
    :type f: Callable
    :return: A response based on if the key is correct or not.
    :rtype: Response
    """

    @wraps(f)
    def check_header(*args, **kwargs):
        auth_header = request.headers.get("Authorization")

        user = db.session.scalar(
            sa.select(Global).where(Global.id == app.config.get("GLOBAL_ID")))

        if user is None or auth_header is None or not user.check_password(auth_header):

            if current_user.is_authenticated:
                logout_user()

                return redirect(url_for(LOGIN_ENDPOINT))

            return create_error_response(UNAUTHORIZED_MSG, UNAUTHORIZED_CODE)

        return f(*args, **kwargs)

    return check_header

@app.route("/login", methods = ["GET", "POST"])
def login() -> Response:
    """
    Endpoint for logging in a user.

    :return: A response based on the uploaded login form.
    :rtype: Response
    """

    if current_user.is_authenticated:
        return redirect(url_for(MAIN_ENDPOINT))

    form = LoginForm()

    if form.validate_on_submit():

        user = db.session.scalar(
            sa.select(Global).where(Global.id == app.config.get("GLOBAL_ID")))

        if user is None or not user.check_password(form.password.data):
            flash("Invalid password.")
            return redirect(url_for(LOGIN_ENDPOINT))

        login_user(user)

        next_page = request.args.get("next")

        if not next_page or urlsplit(next_page).netloc != "":
            next_page = url_for(MAIN_ENDPOINT)

        res = make_response(redirect(next_page))
        res.set_cookie("token", form.password.data)

        return res

    return render_template("login.html", form = form)


@app.route("/", methods = ["GET"])
@app.route("/index", methods = ["GET"])
@login_required
def index() -> Response:
    """
    Endpoint for loading the main page.

    :return: A response holding the main page.
    :rtype: Response
    """

    args_list = []

    args_list.append(CURRENT_ARG)

    debug_on = app.config.get("DEBUG_MODE")

    current_project_request = webhandler.handle_input(webhandler.project_command_name,
                                                      args_list, app.config.get("DEBUG_MODE"))

    curr_project_name = ""

    if current_project_request.result_type is not CommandResultType.FAILURE:
        current_project = current_project_request.payload[0]

        if current_project is not None and current_project.type is not ResponseType.ERROR:
            curr_project_name = current_project.payload.name

    page_theme = request.cookies.get('theme') == 'true'

    return render_template("index.html", curr_project_name = curr_project_name,
                           debug = debug_on, theme=page_theme)


@app.route("/api/favorite/<int:v_id>", methods = ["POST"])
@check_auth_token
def favorite(v_id) -> Response:
    """
    Endpoint for favorising a version.

    :return: A response with an error or the modified data from the database.
    :rtype: Response
    """

    args_list = []

    args_list.extend((FAVORITE_ARG, str(v_id)))

    favorite_version_request = webhandler.handle_input(webhandler.version_command_name,
                                                       args_list, app.config.get("DEBUG_MODE"))

    if favorite_version_request.result_type is CommandResultType.FAILURE:
        return create_error_response(favorite_version_request.message, COMMAND_ERROR_CODE)

    favorited_version = favorite_version_request.payload[0]

    if favorited_version.type is ResponseType.ERROR:
        return create_error_response(favorited_version.message, COMMAND_ERROR_CODE)

    return create_payload_response(favorited_version.payload.get_model_protected(),
                                   favorited_version.message)


@app.route("/api/delete/<string:_type>/<string:identifier>", methods = ["POST"])
@check_auth_token
def delete(_type, identifier) -> Response:
    """
    Endpoint for deleting either a version or project.

    :return: A response with an error or the modified data from the database.
    :rtype: Response
    """

    _type = string_to_bool(_type)

    delete_request = None
    args_list = []

    args_list.extend((DELETE_ARG, identifier))

    debug_on = app.config.get("DEBUG_MODE")

    if _type:
        args_list.append(DELETE_SKIP_CONF_ARG)

        delete_request = webhandler.handle_input(webhandler.version_command_name,
                                                 args_list, debug_on)
    else:
        delete_request = webhandler.handle_input(webhandler.project_command_name,
                                                 args_list, debug_on)

    if delete_request.result_type is CommandResultType.FAILURE:
        return create_error_response(delete_request.message, COMMAND_ERROR_CODE)

    deleted_entity = delete_request.payload[0]

    if deleted_entity.type is ResponseType.ERROR:
        return create_error_response(deleted_entity.message, COMMAND_ERROR_CODE)

    return create_payload_response(deleted_entity.payload.get_model_protected(),
                                   deleted_entity.message)


@app.route("/api/new/<string:name>", methods = ["POST"])
@check_auth_token
def new(name) -> Response:
    """
    Endpoint for creating a new project.

    :return: A response with an error or the modified data from the database.
    :rtype: Response
    """

    args_list = []

    args_list.extend((NEW_ARG, name))

    add_project_request = webhandler.handle_input(webhandler.project_command_name,
                                                  args_list, app.config.get("DEBUG_MODE"))

    if add_project_request.result_type is CommandResultType.FAILURE:
        return create_error_response(add_project_request.message, COMMAND_ERROR_CODE)

    added_project = add_project_request.payload[0]

    if added_project.type is ResponseType.ERROR:
        return create_error_response(added_project.message, COMMAND_ERROR_CODE)

    return create_payload_response(added_project.payload.get_model_protected(),
                                   added_project.message)


@app.route("/api/select/<string:name>", methods = ["GET"])
@check_auth_token
def select(name) -> Response:
    """
    Endpoint for selecting a specific project.

    :return: A response with an error or the modified data from the database.
    :rtype: Response
    """

    args_list = []

    args_list.extend((OPEN_ARG, name))

    select_project_request = webhandler.handle_input(webhandler.project_command_name,
                                                     args_list, app.config.get("DEBUG_MODE"))

    if select_project_request.result_type is CommandResultType.FAILURE:
        return create_error_response(select_project_request.message, COMMAND_ERROR_CODE)

    selected_project = select_project_request.payload[0]

    if selected_project.type is ResponseType.ERROR:
        return create_error_response(selected_project.message, COMMAND_ERROR_CODE)

    return create_payload_response(selected_project.payload.get_model_protected(),
                                   selected_project.message)


@app.route("/api/config/limit/<int:num>", methods = ["POST"])
@check_auth_token
def limit(num) -> Response:
    """
    Endpoint for changing the version limit.

    :return: A response with an error or the modified data from the database.
    :rtype: Response
    """

    args_list = []

    args_list.extend((LIMIT_ARG, str(num)))

    change_limit_request = webhandler.handle_input(webhandler.project_command_name,
                                                   args_list, app.config.get("DEBUG_MODE"))

    if change_limit_request.result_type is CommandResultType.FAILURE:
        return create_error_response(change_limit_request.message, COMMAND_ERROR_CODE)

    changed_config = change_limit_request.payload[0]

    if changed_config.type is ResponseType.ERROR:
        return create_error_response(changed_config.message, COMMAND_ERROR_CODE)

    return create_payload_response(changed_config.payload.get_model_protected(),
                                   changed_config.message)


@app.route("/api/all/<string:_type>", methods = ["GET"])
@app.route("/api/all/<string:_type>/<int:p_id>", methods = ["GET"])
@check_auth_token
def all(_type, p_id = None) -> Response:
    """
    Endpoint for receiving ALL projects OR versions.

    :return: A response with an error or the data from the database.
    :rtype: Response
    """

    _type = string_to_bool(_type)

    get_all_request = None
    args_list = []

    args_list.append(GET_ALL_ARG)

    debug_on = app.config.get("DEBUG_MODE")

    if _type:
        if p_id is not None:
            args_list.append(str(p_id))

        get_all_request = webhandler.handle_input(webhandler.version_command_name,
                                                  args_list, debug_on)

    else:
        get_all_request = webhandler.handle_input(webhandler.project_command_name,
                                                  args_list, debug_on)

    if get_all_request.result_type is CommandResultType.FAILURE:
        return create_error_response(get_all_request.message, COMMAND_ERROR_CODE)

    all_entities = get_all_request.payload

    res = []

    res_message = None

    for entity in all_entities:
        if entity.type is ResponseType.ERROR:
            return create_error_response(entity.message, COMMAND_ERROR_CODE)

        return_payload = entity.payload.get_model_protected()

        res_message = entity.message

        if hasattr(entity, "versions") and entity.versions is not None:
            return_payload["versions"] = []

            for version in entity.versions:
                return_payload["versions"].append(version.get_model_protected())

        res.append(return_payload)

    return create_payload_response(res, res_message)


@app.route("/api/config", methods = ["GET"])
@check_auth_token
def config() -> Response:
    """
    Endpoint for receiving current config (settings).

    :return: A response with an error or the data from the database.
    :rtype: Response
    """

    args_list = []

    args_list.append(CONFIG_ARG)

    get_config_request = webhandler.handle_input(webhandler.project_command_name,
                                                 args_list, app.config.get("DEBUG_MODE"))

    if get_config_request.result_type is CommandResultType.FAILURE:
        return create_error_response(get_config_request.message, COMMAND_ERROR_CODE)

    current_config = get_config_request.payload[0]

    if current_config.type is ResponseType.ERROR:
        return create_error_response(current_config.message, COMMAND_ERROR_CODE)

    return create_payload_response(current_config.payload.get_model_protected(),
                                   current_config.message)


@app.route("/api/current", methods = ["GET"])
@check_auth_token
def current() -> Response:
    """
    Endpoint for receiving all data of the current project.

    :return: A response with an error or the data from the database.
    :rtype: Response
    """

    args_list = []

    args_list.append(CURRENT_ARG)

    get_current_project_request = webhandler.handle_input(webhandler.project_command_name,
                                                          args_list, app.config.get("DEBUG_MODE"))

    if get_current_project_request.result_type is CommandResultType.FAILURE:
        return create_error_response(get_current_project_request.message, COMMAND_ERROR_CODE)

    current_project = get_current_project_request.payload[0]

    if current_project.type is ResponseType.ERROR:
        return create_error_response(current_project.message, COMMAND_ERROR_CODE)

    return create_payload_response(current_project.payload.get_model_protected(),
                                   current_project.message)


@app.route("/api/parse/<string:_run>", methods = ["POST"])
@check_auth_token
def parse(_run) -> Response:
    """
    Endpoint for parsing a certain file.

    :return: A response with an error or the modified data from the database.
    :rtype: Response
    """

    _run = string_to_bool(_run)

    uploaded_file = request.files.get("file")

    file_filename = secure_filename(uploaded_file.filename)
    file_parts = os.path.splitext(file_filename)

    file_name = file_parts[0]

    if not re.match(app.config.get("VALID_FILE_NAME_FORMAT"), file_name):
        return create_error_response(INVALID_FILE_NAME_FORMAT_MSG, COMMAND_ERROR_CODE)

    file_ext = file_parts[1]

    if file_ext not in app.config.get("VALID_FILE_EXTENSIONS"):
        return create_error_response(INVALID_FILE_EXTENSION_MSG, COMMAND_ERROR_CODE)

    abs_path = os.path.join(basedir, app.config.get("RELATIVE_FILE_PATH"))
    file_path = os.path.join(abs_path, file_filename)

    if os.path.isfile(file_path):
        new_name = increase_file_number(file_name, abs_path, file_ext)
        new_filename = new_name + file_ext

        file_path = os.path.join(abs_path, new_filename)

    with open(file_path, "wb") as file:
        uploaded_file.save(file)

    parse_request = None
    args_list = []

    args_list.append(file_path)

    parse_request = webhandler.handle_input(webhandler.parse_command_name,
                                            args_list, app.config.get("DEBUG_MODE"))

    os.remove(file_path)

    if parse_request.result_type is CommandResultType.FAILURE:
        return create_error_response(parse_request.message, COMMAND_ERROR_CODE)

    added_version = parse_request.payload[0]
    removed_version = None

    if (len(parse_request.payload) > 1 and parse_request.payload[1] is not None
            and parse_request.payload[1].payload is not None):
        removed_version = parse_request.payload[1].payload.get_model_protected()

    if added_version.type is ResponseType.ERROR:
        return create_error_response(added_version.message, COMMAND_ERROR_CODE)

    if _run and app.config["RUNNING_THREAD"] is None:
        args_list = []

        args_list.extend((RUN_ARG, str(added_version.payload.v_id)))

        app.config["RUNNING_THREAD"] = Thread(
            target = webhandler.handle_input,
            args = (webhandler.version_command_name, args_list, app.config.get("DEBUG_MODE"))
        )

        app.config["RUNNING_THREAD"].start()

    return create_payload_response(
        [added_version.payload.get_model_protected(), removed_version],
        added_version.message
    )


@app.route("/api/run/<int:v_id>", methods = ["POST"])
@check_auth_token
def run(v_id) -> Response:
    """
    Endpoint for running a certain version.

    :return: A response with an error or no data.
    :rtype: Response
    """

    args_list = []

    args_list.append(GET_ALL_ARG)

    get_all_request = webhandler.handle_input(webhandler.version_command_name,
                                                args_list, app.config.get("DEBUG_MODE"))

    if get_all_request.result_type is CommandResultType.FAILURE:
        return create_error_response(get_all_request.message, COMMAND_ERROR_CODE)

    versions = get_all_request.payload
    version_found = False

    for version in versions:
        if version.type is ResponseType.ERROR:
            return create_error_response(version.message, COMMAND_ERROR_CODE)

        if version.payload.v_id == v_id:
            version_found = True
            break

    if not version_found:
        return create_error_response("Cannot run a version with an invalid id.", COMMAND_ERROR_CODE)

    args_list = []

    args_list.extend((RUN_ARG, str(v_id)))

    res_message = "The controller is already running."

    if app.config["RUNNING_THREAD"] is None:
        app.config["RUNNING_THREAD"] = Thread(
            target = webhandler.handle_input,
            args = (webhandler.version_command_name, args_list, app.config.get("DEBUG_MODE"))
        )

        app.config["RUNNING_THREAD"].start()

        res_message = f"Started running the version with the id: {v_id}."

        return create_payload_response(None, res_message)

    return create_error_response(res_message, COMMAND_ERROR_CODE)


@app.route("/api/stop", methods = ["POST"])
@check_auth_token
def stop() -> Response:
    """
    Endpoint for stopping a running version.

    :return: A response with an error or no data.
    :rtype: Response
    """

    curr_thread = app.config.get("RUNNING_THREAD")

    if curr_thread is not None and curr_thread.is_alive():
        app.config["RUNNING_THREAD"] = None

    Executor.stop()

    args_list = []

    args_list.append(STOP_ARG)

    stop_controller_request = webhandler.handle_input(webhandler.version_command_name,
                                                          args_list, app.config.get("DEBUG_MODE"))

    if stop_controller_request.result_type is CommandResultType.FAILURE:
        return create_error_response(stop_controller_request.message, COMMAND_ERROR_CODE)

    return create_payload_response(None, stop_controller_request.message)


@app.route("/api/help", methods = ["GET"])
def help() -> Response:
    """
    Endpoint for opening or receiving the manual pdf file.

    :return: A response with the manual file.
    :rtype: Response
    """

    return send_file(os.path.join(basedir, HANDBUCH_RELATIVE_PATH + HANDBUCH_FILE_NAME),
                     download_name = HANDBUCH_FILE_NAME)


@app.route("/api/update", methods = ["POST"])
def update() -> Response:
    """
    Endpoint for receiving updates regarding the currently running automata from the Controller

    :return: A response with no data.
    :rtype: Response
    """

    update_msg = request.get_json()["body"]["message"]

    update_socket(update_msg)

    return create_payload_response(None, None)


@app.route("/api/debug", methods = ["POST"])
@check_auth_token
def debug() -> Response:
    """
    Endpoint for toggling on and off the debug mode

    :return: A response with no data.
    :rtype: Response
    """

    app.config["DEBUG_MODE"] = not app.config.get("DEBUG_MODE")

    return create_payload_response(None, None)

@app.route('/favicon.ico')
def favicon() -> Response:
    """
    Endpoint for retrieving the icon of the website.

    :return: The icon of the website.
    :rtype: Response
    """

    return send_from_directory(os.path.join(app.root_path, ICON_RELATIVE_PATH),
                               ICON_FILE_NAME, mimetype='image/vnd.microsoft.icon')
